from __future__ import annotations

import re
from importlib.metadata import version
from textwrap import dedent

import pytest


@pytest.fixture
def flake8_path(flake8_path):
    (flake8_path / "setup.cfg").write_text(
        dedent(
            """\
            [flake8]
            select = C4
            """
        )
    )
    yield flake8_path


def test_version(flake8_path):
    result = flake8_path.run_flake8(["--version"])
    version_regex = r"flake8-comprehensions:( )*" + version("flake8-comprehensions")
    unwrapped = "".join(result.out_lines)
    assert re.search(version_regex, unwrapped)


@pytest.mark.parametrize(
    "code",
    [
        "foo = [x + 1 for x in range(10)]",
    ],
)
def test_C400_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = list(x + 1 for x in range(10))",
            [
                "./example.py:1:7: C400 Unnecessary generator - rewrite as a list "
                + "comprehension."
            ],
        ),
        (
            """\
            foobar = list(
                str(x)
                for x
                in range(10)
            )
            """,
            [
                "./example.py:1:10: C400 Unnecessary generator - rewrite as a list "
                + "comprehension."
            ],
        ),
    ],
)
def test_C400_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = {x + 1 for x in range(10)}",
    ],
)
def test_C401_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = set(x + 1 for x in range(10))",
            [
                "./example.py:1:7: C401 Unnecessary generator - rewrite as a set "
                + "comprehension."
            ],
        ),
        (
            """\
            foobar = set(
                str(x) for x
                in range(10)
            )
            """,
            [
                "./example.py:1:10: C401 Unnecessary generator - rewrite as a set "
                + "comprehension."
            ],
        ),
    ],
)
def test_C401_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = {x: str(x) for x in range(10)}",
        """\
        foo = ['a=1', 'b=2', 'c=3']
        dict(pair.split('=') for pair in foo)
        """,
        """\
        foo = [('a', 1), ('b', 2), ('c', 3)]
        dict(pair for pair in foo if pair[1] % 2 == 0)
        """,
        # Previously a false positive:
        "dict(((x, str(x)) for x in range(10)), c=1)",
    ],
)
def test_C402_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = dict((x, str(x)) for x in range(10))",
            [
                "./example.py:1:7: C402 Unnecessary generator - rewrite as a dict "
                + "comprehension."
            ],
        ),
        (
            """\
            foobar = dict(
                (x, str(x))
                for x
                in range(10)
            )
            """,
            [
                "./example.py:1:10: C402 Unnecessary generator - rewrite as a dict "
                + "comprehension."
            ],
        ),
    ],
)
def test_C402_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = {x + 1 for x in range(10)}",
    ],
)
def test_C403_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = set([x + 1 for x in range(10)])",
            [
                "./example.py:1:7: C403 Unnecessary list comprehension - rewrite as a "
                + "set comprehension."
            ],
        ),
    ],
)
def test_C403_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = {x: x for x in range(10)}",
        # Previously a false positive:
        "foo = dict([x.split('=') for x in ['a=1', 'b=2']])",
        # Previously a false positive:
        "dict([(x, x) for x in range(10)], y=2)",
    ],
)
def test_C404_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = dict([(x, x) for x in range(10)])",
            [
                "./example.py:1:7: C404 Unnecessary list comprehension - rewrite as a "
                + "dict comprehension."
            ],
        ),
    ],
)
def test_C404_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = set(range)",
    ],
)
def test_C405_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = set([])",
            [
                "./example.py:1:7: C405 Unnecessary list literal - rewrite as a set "
                + "literal."
            ],
        ),
        (
            "foo = set([1])",
            [
                "./example.py:1:7: C405 Unnecessary list literal - rewrite as a set "
                + "literal."
            ],
        ),
        (
            "foo = set(())",
            [
                "./example.py:1:7: C405 Unnecessary tuple literal - rewrite as a set "
                + "literal."
            ],
        ),
        (
            "foo = set((1,))",
            [
                "./example.py:1:7: C405 Unnecessary tuple literal - rewrite as a set "
                + "literal."
            ],
        ),
    ],
)
def test_C405_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = dict(range)",
        "something = (1, 2); dict([something])",
        "dict([(1,)])",
    ],
)
def test_C406_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = dict([])",
            [
                "./example.py:1:7: C406 Unnecessary list literal - rewrite as a dict "
                + "literal."
            ],
        ),
        (
            "foo = dict([(1, 2)])",
            [
                "./example.py:1:7: C406 Unnecessary list literal - rewrite as a dict "
                + "literal."
            ],
        ),
        (
            "foo = dict(())",
            [
                "./example.py:1:7: C406 Unnecessary tuple literal - rewrite as a dict "
                + "literal."
            ],
        ),
        (
            "foo = dict(((1, 2),))",
            [
                "./example.py:1:7: C406 Unnecessary tuple literal - rewrite as a dict "
                + "literal."
            ],
        ),
    ],
)
def test_C406_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "()",
        "[]",
        "{}",
        "set()",
        """\
        foo = [('foo', 2)]
        dict(foo)
        """,
        """\
        foo = {}
        dict(bar=1, **foo)
        """,
        """\
        foo = [1, 2]
        list(foo)
        """,
        """\
        foo = [1, 2]
        list(*foo)
        """,
    ],
)
def test_C408_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "tuple()",
            ["./example.py:1:1: C408 Unnecessary tuple call - rewrite as a literal."],
        ),
        (
            "list()",
            ["./example.py:1:1: C408 Unnecessary list call - rewrite as a literal."],
        ),
        (
            "dict()",
            ["./example.py:1:1: C408 Unnecessary dict call - rewrite as a literal."],
        ),
        (
            "dict(a=1)",
            ["./example.py:1:1: C408 Unnecessary dict call - rewrite as a literal."],
        ),
    ],
)
def test_C408_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = tuple(range)",
    ],
)
def test_C409_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = tuple([])",
            [
                "./example.py:1:7: C409 Unnecessary list passed to tuple() - "
                + "rewrite as a tuple literal."
            ],
        ),
        (
            "foo = tuple([1, 2])",
            [
                "./example.py:1:7: C409 Unnecessary list passed to tuple() - "
                + "rewrite as a tuple literal."
            ],
        ),
        (
            "foo = tuple(())",
            [
                "./example.py:1:7: C409 Unnecessary tuple passed to tuple() - remove "
                + "the outer call to tuple()."
            ],
        ),
        (
            "foo = tuple((1, 2))",
            [
                "./example.py:1:7: C409 Unnecessary tuple passed to tuple() - remove "
                + "the outer call to tuple()."
            ],
        ),
    ],
)
def test_C409_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "foo = list(range)",
    ],
)
def test_C410_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "foo = list([])",
            [
                "./example.py:1:7: C410 Unnecessary list passed to list() - remove the "
                + "outer call to list()."
            ],
        ),
        (
            "foo = list([1, 2])",
            [
                "./example.py:1:7: C410 Unnecessary list passed to list() - remove the "
                + "outer call to list()."
            ],
        ),
        (
            "foo = list(())",
            [
                "./example.py:1:7: C410 Unnecessary tuple passed to list() - "
                + "rewrite as a list literal."
            ],
        ),
        (
            "foo = list((1, 2))",
            [
                "./example.py:1:7: C410 Unnecessary tuple passed to list() - "
                + "rewrite as a list literal."
            ],
        ),
    ],
)
def test_C410_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "[x + 1 for x in range(10)]",
    ],
)
def test_C411_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "list([x + 1 for x in range(10)])",
            [
                "./example.py:1:1: C411 Unnecessary list call - remove the outer call "
                + "to list()."
            ],
        ),
    ],
)
def test_C411_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "sorted([2, 3, 1])",
        "sorted([2, 3, 1], reverse=True)",
        "sorted([2, 3, 1], reverse=False)",
        "sorted([2, 3, 1], reverse=0)",
        "sorted([2, 3, 1], reverse=1)",
        "reversed([2, 3, 1])",
    ],
)
def test_C413_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "list(sorted([2, 3, 1]))",
            ["./example.py:1:1: C413 Unnecessary list call around sorted()."],
        ),
        (
            "reversed(sorted([2, 3, 1]))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - use sorted(..., reverse=True)."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=False))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - use sorted(..., reverse=True)."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=True))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - use sorted(..., reverse=False)."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=0))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - use sorted(..., reverse=True)."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=1))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - use sorted(..., reverse=False)."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=bool()))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - toggle reverse argument to sorted()."
            ],
        ),
        (
            "reversed(sorted([2, 3, 1], reverse=not True))",
            [
                "./example.py:1:1: C413 Unnecessary reversed call around sorted()"
                + " - toggle reverse argument to sorted()."
            ],
        ),
    ],
)
def test_C413_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "list(set(a))",
        "tuple(set(a))",
        "sorted(set(a))",
    ],
)
def test_C414_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "list(list(a))",
            ["./example.py:1:1: C414 Unnecessary list call within list()."],
        ),
        (
            "list(tuple(a))",
            ["./example.py:1:1: C414 Unnecessary tuple call within list()."],
        ),
        (
            "tuple(list(a))",
            ["./example.py:1:1: C414 Unnecessary list call within tuple()."],
        ),
        (
            "tuple(tuple(a))",
            ["./example.py:1:1: C414 Unnecessary tuple call within tuple()."],
        ),
        ("set(set(a))", ["./example.py:1:1: C414 Unnecessary set call within set()."]),
        (
            "set(list(a))",
            ["./example.py:1:1: C414 Unnecessary list call within set()."],
        ),
        (
            "set(tuple(a))",
            ["./example.py:1:1: C414 Unnecessary tuple call within set()."],
        ),
        (
            "set(sorted(a))",
            ["./example.py:1:1: C414 Unnecessary sorted call within set()."],
        ),
        (
            "set(sorted(a, reverse=True))",
            ["./example.py:1:1: C414 Unnecessary sorted call within set()."],
        ),
        (
            "set(reversed(a))",
            ["./example.py:1:1: C414 Unnecessary reversed call within set()."],
        ),
        (
            "sorted(list(a))",
            ["./example.py:1:1: C414 Unnecessary list call within sorted()."],
        ),
        (
            "sorted(tuple(a))",
            ["./example.py:1:1: C414 Unnecessary tuple call within sorted()."],
        ),
        (
            "sorted(sorted(a))",
            ["./example.py:1:1: C414 Unnecessary sorted call within sorted()."],
        ),
        (
            "sorted(sorted(a), reverse=True)",
            ["./example.py:1:1: C414 Unnecessary sorted call within sorted()."],
        ),
        (
            "sorted(sorted(a, reverse=True))",
            ["./example.py:1:1: C414 Unnecessary sorted call within sorted()."],
        ),
        (
            "sorted(sorted(a, reverse=True), reverse=True)",
            ["./example.py:1:1: C414 Unnecessary sorted call within sorted()."],
        ),
        (
            "sorted(reversed(a))",
            ["./example.py:1:1: C414 Unnecessary reversed call within sorted()."],
        ),
        (
            "sorted(reversed(a), reverse=True)",
            ["./example.py:1:1: C414 Unnecessary reversed call within sorted()."],
        ),
    ],
)
def test_C414_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "set([2, 3, 1][::1])",
        "sorted([2, 3, 1][::1])",
        "reversed([2, 3, 1][::1])",
    ],
)
def test_C415_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "set([2, 3, 1][::-1])",
            [
                "./example.py:1:1: C415 Unnecessary subscript reversal of iterable "
                + "within set()."
            ],
        ),
        (
            "sorted([2, 3, 1][::-1])",
            [
                "./example.py:1:1: C415 Unnecessary subscript reversal of iterable "
                + "within sorted()."
            ],
        ),
        (
            "sorted([2, 3, 1][::-1], reverse=True)",
            [
                "./example.py:1:1: C415 Unnecessary subscript reversal of iterable "
                + "within sorted()."
            ],
        ),
        (
            "reversed([2, 3, 1][::-1])",
            [
                "./example.py:1:1: C415 Unnecessary subscript reversal of iterable "
                + "within reversed()."
            ],
        ),
    ],
)
def test_C415_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "{x, y for x, y, z in zip('abc', '123', 'def')}",
        "{y: x for x, y in zip('abc', '123')}",
        "{x: y for x, (y,) in zip('a', ('1',))}",
        "{x: z for x, (y,), z in zip('a', ('1',), 'b')}",
        "[str(x) for x in range(5)]",
        "[x + 1 for x in range(5)]",
        "[x for x in range(5) if x % 2]",
        "{str(x) for x in range(5)}",
        "{x + 1 for x in range(5)}",
        "{x for x in range(5) if x % 2}",
        """\
        async def foo():
            [x async for x in range(5)]
        """,
        """\
        async def foo():
            return {x async for x in range(5)}
        """,
        "[(x, y, 1) for x, y in []]",
        # We can't assume unpacking came from tuples:
        "[(x, y) for x, y in zip('abc', '123')]",
        "[(x, y) for (x, y) in zip('abc', '123')]",
        "{(x, y) for x, y in zip('abc', '123')}",
        "{(x, y) for (x, y) in zip('abc', '123')}",
    ],
)
def test_C416_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "{x: y for x, y in zip(range(5), range(5))}",
            [
                "./example.py:1:1: C416 Unnecessary dict comprehension - "
                + "rewrite using dict().",
            ],
        ),
        (
            "{x: y for (x, y) in zip(range(5), range(5))}",
            [
                "./example.py:1:1: C416 Unnecessary dict comprehension - "
                + "rewrite using dict().",
            ],
        ),
        (
            "[x for x in range(5)]",
            [
                "./example.py:1:1: C416 Unnecessary "
                + "list comprehension - rewrite using list()."
            ],
        ),
        (
            "{x for x in range(5)}",
            [
                "./example.py:1:1: C416 Unnecessary set comprehension - "
                + "rewrite using set().",
            ],
        ),
    ],
)
def test_C416_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "map()",
        "map(str, numbers)",
        "list(map())",
        "list(map(str, numbers))",
        "set(map(f, items))",
        "dict(map(enumerate, values))",
        "dict(map(lambda v: data[v], values))",
    ],
)
def test_C417_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "map(lambda x: x * 2, iterable)",
            [
                "./example.py:1:1: C417 Unnecessary use of map - "
                + "use a generator expression instead.",
            ],
        ),
        (
            "list(map(lambda x: x * 2, iterable))",
            [
                "./example.py:1:1: C417 Unnecessary use of map - "
                + "use a list comprehension instead.",
            ],
        ),
        (
            "set(map(lambda num: num % 2 == 0, nums))",
            [
                "./example.py:1:1: C417 Unnecessary use of map - "
                + "use a set comprehension instead.",
            ],
        ),
        (
            "dict(map(lambda v: (v, v ** 2), values))",
            [
                "./example.py:1:1: C417 Unnecessary use of map - "
                "use a dict comprehension instead.",
            ],
        ),
    ],
)
def test_C417_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "dict({}, a=1)",
        "dict({x: [] for x in range(1)}, a=1)",
    ],
)
def test_C418_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "dict({})",
            [
                "./example.py:1:1: C418 Unnecessary dict passed to dict() - "
                + "remove the outer call to dict()."
            ],
        ),
        (
            "dict({'a': 1})",
            [
                "./example.py:1:1: C418 Unnecessary dict passed to dict() - "
                + "remove the outer call to dict()."
            ],
        ),
        (
            "dict({'x': 1 for x in range(10)})",
            [
                "./example.py:1:1: C418 Unnecessary dict comprehension passed "
                + "to dict() - remove the outer call to dict()."
            ],
        ),
    ],
)
def test_C418_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "any(num == 3 for num in range(5))",
        "all(num == 3 for num in range(5))",
    ],
)
def test_C419_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "any([num == 3 for num in range(5)])",
            [
                "./example.py:1:1: C419 Unnecessary list comprehension passed "
                + "to any() prevents short-circuiting - rewrite as a generator."
            ],
        ),
        (
            "all([num == 3 for num in range(5)])",
            [
                "./example.py:1:1: C419 Unnecessary list comprehension passed "
                + "to all() prevents short-circuiting - rewrite as a generator."
            ],
        ),
    ],
)
def test_C419_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures


@pytest.mark.parametrize(
    "code",
    [
        "{elt: elt * 2 for elt in range(5)}",
        "{elt: [] for elt in foo}",
        "{elt: {1, 2, 3} for elt in ['a', 'b', 'c']}",
        "{elt: some_func() for elt in ['a', 'b', 'c']}",
        "{elt: SomeClass() for elt in ['a', 'b', 'c']}",
    ],
)
def test_C420_pass(code, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == []


@pytest.mark.parametrize(
    "code,failures",
    [
        (
            "{elt: None for elt in range(5)}",
            [
                "./example.py:1:1: C420 Unnecessary dict comprehension - "
                + "rewrite using dict.fromkeys()."
            ],
        ),
        (
            "{elt: 1 for elt in foo}",
            [
                "./example.py:1:1: C420 Unnecessary dict comprehension - "
                + "rewrite using dict.fromkeys()."
            ],
        ),
        (
            "{elt: 'value' for elt in ['a', 'b', 'c']}",
            [
                "./example.py:1:1: C420 Unnecessary dict comprehension - "
                + "rewrite using dict.fromkeys()."
            ],
        ),
        (
            "{elt: True for elt in some_func()}",
            [
                "./example.py:1:1: C420 Unnecessary dict comprehension - "
                + "rewrite using dict.fromkeys()."
            ],
        ),
    ],
)
def test_C420_fail(code, failures, flake8_path):
    (flake8_path / "example.py").write_text(dedent(code))
    result = flake8_path.run_flake8()
    assert result.out_lines == failures



from __future__ import annotations

import ast
from collections.abc import Generator
from importlib.metadata import version
from typing import Any


class ComprehensionChecker:
    """
    Flake8 plugin to help you write better list/set/dict comprehensions.
    """

    name = "flake8-comprehensions"
    version = version("flake8-comprehensions")

    __slots__ = ("tree",)

    def __init__(self, tree: ast.AST) -> None:
        self.tree = tree

    messages = {
        "C400": "C400 Unnecessary generator - rewrite as a list comprehension.",
        "C401": "C401 Unnecessary generator - rewrite as a set comprehension.",
        "C402": "C402 Unnecessary generator - rewrite as a dict comprehension.",
        "C403": "C403 Unnecessary list comprehension - rewrite as a set comprehension.",
        "C404": (
            "C404 Unnecessary list comprehension - rewrite as a dict comprehension."
        ),
        "C405": "C405 Unnecessary {type} literal - ",
        "C406": "C406 Unnecessary {type} literal - ",
        "C408": "C408 Unnecessary {type} call - rewrite as a literal.",
        "C409": "C409 Unnecessary {type} passed to tuple() - ",
        "C410": "C410 Unnecessary {type} passed to list() - ",
        "C411": "C411 Unnecessary list call - remove the outer call to list().",
        "C413": "C413 Unnecessary {outer} call around {inner}(){remediation}.",
        "C414": "C414 Unnecessary {inner} call within {outer}().",
        "C415": "C415 Unnecessary subscript reversal of iterable within {func}().",
        "C416": "C416 Unnecessary {type} comprehension - rewrite using {type}().",
        "C417": "C417 Unnecessary use of map - use a {comp} instead.",
        "C418": (
            "C418 Unnecessary {type} passed to dict() - "
            + "remove the outer call to dict()."
        ),
        "C419": (
            "C419 Unnecessary list comprehension passed to {func}() prevents "
            + "short-circuiting - rewrite as a generator."
        ),
        "C420": (
            "C420 Unnecessary {type} comprehension - rewrite using dict.fromkeys()."
        ),
    }

    def run(self) -> Generator[tuple[int, int, str, type[Any]]]:
        # Stores previously seen map() nodes, to avoid raising C417 on it twice.
        visited_map_calls: set[ast.Call] = set()

        for node in ast.walk(self.tree):
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Name):
                num_positional_args = len(node.args)
                num_keyword_args = len(node.keywords)

                if (
                    num_positional_args == 1
                    and isinstance(node.args[0], ast.GeneratorExp)
                    and node.func.id in ("list", "set")
                ):
                    msg_key = {"list": "C400", "set": "C401"}[node.func.id]
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages[msg_key],
                        type(self),
                    )

                elif (
                    num_positional_args == 1
                    and node.func.id == "dict"
                    and len(node.keywords) == 0
                    and isinstance(node.args[0], (ast.GeneratorExp, ast.ListComp))
                    and isinstance(node.args[0].elt, ast.Tuple)
                    and len(node.args[0].elt.elts) == 2
                ):
                    if isinstance(node.args[0], ast.GeneratorExp):
                        msg = "C402"
                    else:
                        msg = "C404"
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages[msg],
                        type(self),
                    )

                elif (
                    num_positional_args == 1
                    and isinstance(node.args[0], ast.ListComp)
                    and node.func.id in ("list", "set", "any", "all")
                ):
                    msg_key = {
                        "list": "C411",
                        "set": "C403",
                        "any": "C419",
                        "all": "C419",
                    }[node.func.id]
                    msg = self.messages[msg_key].format(func=node.func.id)
                    yield (
                        node.lineno,
                        node.col_offset,
                        msg,
                        type(self),
                    )

                elif num_positional_args == 1 and (
                    isinstance(node.args[0], ast.Tuple)
                    and node.func.id == "tuple"
                    or isinstance(node.args[0], ast.List)
                    and node.func.id == "list"
                ):
                    suffix = "remove the outer call to {func}()."
                    msg_key = {"tuple": "C409", "list": "C410"}[node.func.id]
                    msg = self.messages[msg_key] + suffix
                    yield (
                        node.lineno,
                        node.col_offset,
                        msg.format(
                            type=type(node.args[0]).__name__.lower(), func=node.func.id
                        ),
                        type(self),
                    )

                elif (
                    num_positional_args == 1
                    and num_keyword_args == 0
                    and isinstance(node.args[0], (ast.Dict, ast.DictComp))
                    and node.func.id == "dict"
                ):
                    if isinstance(node.args[0], ast.Dict):
                        type_ = "dict"
                    else:
                        type_ = "dict comprehension"
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C418"].format(type=type_),
                        type(self),
                    )

                elif (
                    num_positional_args == 1
                    and isinstance(node.args[0], (ast.Tuple, ast.List))
                    and (
                        node.func.id in ("tuple", "list", "set")
                        or (
                            node.func.id == "dict"
                            and all(
                                isinstance(i, ast.Tuple) and len(i.elts) == 2
                                for i in node.args[0].elts
                            )
                        )
                    )
                ):
                    suffix = "rewrite as a {func} literal."
                    msg_key = {
                        "tuple": "C409",
                        "list": "C410",
                        "set": "C405",
                        "dict": "C406",
                    }[node.func.id]
                    msg = self.messages[msg_key] + suffix
                    yield (
                        node.lineno,
                        node.col_offset,
                        msg.format(
                            type=type(node.args[0]).__name__.lower(), func=node.func.id
                        ),
                        type(self),
                    )

                elif (
                    num_positional_args == 0
                    and not has_star_args(node)
                    and not has_double_star_args(node)
                    and node.func.id == "dict"
                ):
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C408"].format(type=node.func.id),
                        type(self),
                    )

                elif (
                    num_positional_args == 0
                    and num_keyword_args == 0
                    and node.func.id in ("tuple", "list")
                ):
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C408"].format(type=node.func.id),
                        type(self),
                    )

                elif (
                    node.func.id in {"list", "reversed"}
                    and num_positional_args > 0
                    and isinstance(node.args[0], ast.Call)
                    and isinstance(node.args[0].func, ast.Name)
                    and node.args[0].func.id == "sorted"
                ):
                    remediation = ""
                    if node.func.id == "reversed":
                        reverse_flag_value: bool | None = False
                        for keyword in node.args[0].keywords:
                            if keyword.arg != "reverse":
                                continue
                            if isinstance(keyword.value, ast.NameConstant):
                                reverse_flag_value = keyword.value.value
                            elif isinstance(keyword.value, ast.Num):
                                reverse_flag_value = bool(keyword.value.n)
                            else:
                                # Complex value
                                reverse_flag_value = None

                        if reverse_flag_value is None:
                            remediation = " - toggle reverse argument to sorted()"
                        else:
                            remediation = " - use sorted(..., reverse={!r})".format(
                                not reverse_flag_value
                            )

                    msg = self.messages["C413"].format(
                        inner=node.args[0].func.id,
                        outer=node.func.id,
                        remediation=remediation,
                    )
                    yield (
                        node.lineno,
                        node.col_offset,
                        msg,
                        type(self),
                    )

                elif (
                    num_positional_args > 0
                    and isinstance(node.args[0], ast.Call)
                    and isinstance(node.args[0].func, ast.Name)
                    and (
                        (
                            node.func.id in {"set", "sorted"}
                            and node.args[0].func.id
                            in {"list", "reversed", "sorted", "tuple"}
                        )
                        or (
                            node.func.id in {"list", "tuple"}
                            and node.args[0].func.id in {"list", "tuple"}
                        )
                        or (node.func.id == "set" and node.args[0].func.id == "set")
                    )
                ):
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C414"].format(
                            inner=node.args[0].func.id, outer=node.func.id
                        ),
                        type(self),
                    )

                elif (
                    node.func.id in {"reversed", "set", "sorted"}
                    and num_positional_args > 0
                    and isinstance(node.args[0], ast.Subscript)
                    and isinstance(node.args[0].slice, ast.Slice)
                    and node.args[0].slice.lower is None
                    and node.args[0].slice.upper is None
                    and isinstance(node.args[0].slice.step, ast.UnaryOp)
                    and isinstance(node.args[0].slice.step.op, ast.USub)
                    and isinstance(node.args[0].slice.step.operand, ast.Num)
                    and node.args[0].slice.step.operand.n == 1
                ):
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C415"].format(func=node.func.id),
                        type(self),
                    )

                elif (
                    node.func.id == "map"
                    and node not in visited_map_calls
                    and len(node.args) == 2
                    and isinstance(node.args[0], ast.Lambda)
                ):
                    yield (
                        node.lineno,
                        node.col_offset,
                        self.messages["C417"].format(comp="generator expression"),
                        type(self),
                    )

                elif (
                    node.func.id in ("list", "set", "dict")
                    and len(node.args) == 1
                    and isinstance(node.args[0], ast.Call)
                    and isinstance(node.args[0].func, ast.Name)
                    and node.args[0].func.id == "map"
                    and len(node.args[0].args) == 2
                    and isinstance(node.args[0].args[0], ast.Lambda)
                ):
                    # To avoid raising C417 on the map() call inside the list/set/dict.
                    map_call = node.args[0]
                    visited_map_calls.add(map_call)

                    rewriteable = True
                    if node.func.id == "dict":
                        # For the generator expression to be rewriteable as a
                        # dict comprehension, its lambda must return a 2-tuple.
                        lambda_node = node.args[0].args[0]
                        if (
                            not isinstance(lambda_node.body, (ast.List, ast.Tuple))
                            or len(lambda_node.body.elts) != 2
                        ):
                            rewriteable = False

                    if rewriteable:
                        comprehension_type = f"{node.func.id} comprehension"
                        yield (
                            node.lineno,
                            node.col_offset,
                            self.messages["C417"].format(comp=comprehension_type),
                            type(self),
                        )

            elif isinstance(node, (ast.DictComp, ast.ListComp, ast.SetComp)):
                if (
                    len(node.generators) == 1
                    and not node.generators[0].ifs
                    and not node.generators[0].is_async
                ):
                    if (
                        isinstance(node, (ast.ListComp, ast.SetComp))
                        and isinstance(node.elt, ast.Name)
                        and isinstance(node.generators[0].target, ast.Name)
                        and node.elt.id == node.generators[0].target.id
                    ) or (
                        isinstance(node, ast.DictComp)
                        and isinstance(node.key, ast.Name)
                        and isinstance(node.value, ast.Name)
                        and isinstance(node.generators[0].target, ast.Tuple)
                        and len(node.generators[0].target.elts) == 2
                        and isinstance(node.generators[0].target.elts[0], ast.Name)
                        and node.generators[0].target.elts[0].id == node.key.id
                        and isinstance(node.generators[0].target.elts[1], ast.Name)
                        and node.generators[0].target.elts[1].id == node.value.id
                    ):
                        yield (
                            node.lineno,
                            node.col_offset,
                            self.messages["C416"].format(
                                type=comp_type[node.__class__]
                            ),
                            type(self),
                        )

                    elif (
                        isinstance(node, ast.DictComp)
                        and isinstance(node.key, ast.Name)
                        and isinstance(node.value, ast.Constant)
                        and isinstance(node.generators[0].target, ast.Name)
                        and node.key.id == node.generators[0].target.id
                    ):
                        yield (
                            node.lineno,
                            node.col_offset,
                            self.messages["C420"].format(
                                type=comp_type[node.__class__]
                            ),
                            type(self),
                        )


def has_star_args(call_node: ast.Call) -> bool:
    return any(isinstance(a, ast.Starred) for a in call_node.args)


def has_double_star_args(call_node: ast.Call) -> bool:
    return any(k.arg is None for k in call_node.keywords)


comp_type = {
    ast.DictComp: "dict",
    ast.ListComp: "list",
    ast.SetComp: "set",
}



#!/usr/bin/env python
from __future__ import annotations

import os
import subprocess
import sys
from functools import partial
from pathlib import Path

if __name__ == "__main__":
    os.chdir(Path(__file__).parent)
    common_args = [
        "uv",
        "pip",
        "compile",
        "--quiet",
        "--generate-hashes",
        "requirements.in",
        *sys.argv[1:],
    ]
    run = partial(subprocess.run, check=True)
    run([*common_args, "--python", "3.9", "--output-file", "py39.txt"])
    run([*common_args, "--python", "3.10", "--output-file", "py310.txt"])
    run([*common_args, "--python", "3.11", "--output-file", "py311.txt"])
    run([*common_args, "--python", "3.12", "--output-file", "py312.txt"])
    run([*common_args, "--python", "3.13", "--output-file", "py313.txt"])
